/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 03/07/22.
//

#ifndef FGGL_ASSETS_LOADER_HPP
#define FGGL_ASSETS_LOADER_HPP

#include <map>
#include <string>
#include <memory>
#include <functional>
#include <queue>
#include <variant>

#include "fggl/assets/types.hpp"
#include "fggl/assets/packed/adapter.hpp"
#include "fggl/data/storage.hpp"

namespace fggl::assets {

	enum class LoadType {
		DIRECT, // given pointer to persistent memory
		STAGED, // given pointer to temp memory
		PATH // given filesystem::path
	};

	struct ResourceRequest {
		AssetGUID m_guid;
		AssetTypeID m_type;
	};


	class Loader {
		public:
			constexpr const static modules::ModuleService service = modules::make_service("fggl::assets::Loader");

			explicit inline Loader(data::Storage *storage, CheckinAdapted *checkin) : m_storage(storage), m_checkin(checkin) {}

			explicit Loader(Loader *parent, data::Storage *storage) : m_parent(parent), m_storage(storage) {};

			// no move, no copy.
			Loader(const Loader &) = delete;
			Loader &operator=(const Loader &) = delete;
			Loader(Loader &&) = delete;
			Loader &operator=(Loader &&) = delete;

			inline void setFactory(AssetTypeID type, Checkin fn, LoadType loading = LoadType::DIRECT) {
				m_factories[type] = std::make_pair(fn, loading);
			}

			inline void unsetFactory(AssetTypeID type) {
				m_factories.erase(type);
			}

			inline void request(const AssetGUID &guid, const AssetTypeID &type) {
				m_requests.push(ResourceRequest{guid, type});
			}

			void loadChain(const AssetGUID &guid, const std::string& pack = "core") {
				loadChain( assets::make_asset_id_rt(pack, guid) );
			}

			void loadChain(AssetID asset, void* userPtr = nullptr) {
				if (!m_checkin->exists(asset)) {
					debug::warning("attempted to chain load unknown assetID {}", asset);
					return;
				}

				// FIXME use the fancy dependency resolution algorithm from the module system
				std::vector<AssetID> loadOrder;
				std::stack<AssetID> openSet;
				openSet.push(asset);
				while ( !openSet.empty() ) {
					auto current = openSet.top();
					openSet.pop();

					if ( !m_checkin->exists(current) ) {
						debug::warning("attempted to load chain with unknown assert, {} - abort!", current);
						return;
					}

					// this WILL (probably) break if an asset ends up in the chain twice
					loadOrder.push_back( current );
					const auto& record = m_checkin->find(current);
					for ( const auto& dep : record.m_requires ) {
						openSet.push( dep );
					}
				}

				processChain(loadOrder, userPtr);
			}

			void processChain( std::vector<AssetID>& loadOrder, void* userPtr = nullptr ) {
				if ( loadOrder.empty() ) {
					return;
				}

				debug::info("Starting chain load");
				for ( auto it = loadOrder.rbegin(); it < loadOrder.rend(); ++it  ) {
					debug::info(" CHAIN -> loading {}", *it );
					load( *it, userPtr );
				}
				debug::info("Ended chain loader");
			}

			inline void load(const AssetGUID& guid, const AssetTypeID& type, void* userPtr = nullptr, const std::string& pack = "core" ) {
				auto assetID = assets::make_asset_id_rt(pack, guid);

				// try checkin load first, falling back if it fails
				bool checkinLoad = load( assetID, userPtr );
				if ( !checkinLoad ) {
					debug::info("could not perform checkin load for {} - missing loaders?", guid);

					auto path = m_storage->resolvePath(data::StorageType::Data, guid);
					loadDirect(path, assetID, type, pack, userPtr);
				}
			}

			bool loadDirect( const std::filesystem::path& path, AssetID asset, AssetTypeID assetType, const std::string& pack, void* userPtr = nullptr ){
				// check if we know how to load this asset (using old loaders)
				auto factoryItr = m_factories.find( assetType );
				if ( factoryItr == m_factories.end() ) {
					debug::error("attempted to load asset with unknown type: {}", assetType.get() );
					return false;
				}

				// perform loading
				LoaderContext ctx {
					.pack = pack,
					.packRoot = m_checkin->getPackPath(pack),
					.assetPath = path
				};

				// perform loading
				const auto& [callback, callbackType] = factoryItr->second;
				if ( callbackType == LoadType::PATH ) {
					callback(this, asset, ctx, userPtr);
					return true;
				}

				debug::log(debug::Level::error, "Tried to use old/unsupported loading method");
				return false;
			}

			bool load(const AssetID &assetId, void* userPtr = nullptr) {
				if ( !m_checkin->exists(assetId) ) {
					#ifndef NDEBUG
						debug::warning("asked to load unknown asset: {}", util::guidToString( util::GUID::make(assetId.get()) ) );
					#else
						debug::warning("asked to load unknown asset: {}", assetId.get());
					#endif
					return false;
				}

				const auto& record = m_checkin->find( assetId );
				return loadDirect( record.m_path, assetId, record.m_assetType, record.m_pack, userPtr );
			}

			void progress() {
				if (m_requests.empty()) {
					return;
				}
				auto &request = m_requests.front();
				load(request.m_guid, request.m_type);
				m_requests.pop();
			}

			[[nodiscard]]
			inline bool done() const {
				return m_requests.empty();
			}

			/**
			 * Complete all remaining loading requests, blocking until complete.
			 */
			inline void finish() {
				while (!done()) {
					progress();
				}
			}

		private:
			Loader *m_parent = nullptr;
			using Config = std::pair<Checkin, LoadType>;
			data::Storage *m_storage;
			CheckinAdapted *m_checkin;
			std::queue<ResourceRequest> m_requests;
			std::map<AssetTypeID, Config> m_factories;
	};

} // namespace fggl::assets

#endif //FGGL_ASSETS_LOADER_HPP
